class_name Gun
extends KinematicBody2D


signal on_die
signal on_fragments
signal on_reset_boost
signal spawn_bullet
signal on_damaged

const Variants = preload("res://entity/gun/variants.gd")
const HP = preload("res://entity/hp.gd")

const _RADIUS := 20.0
const _HEIGHT := 20.0

var hp: HP = HP.new()

var _rotate_speed := 1.5
var _bullets_spawn_interval := 0.4
var _bullets_spawn_elasped := 0.0
var _bullets_variant
var _damage := 1.0
var _current_variant
var _current_type
var _fragments_variant
var _fragments_min := 2
var _fragments_max := 5
var _old_rotate_speed: float
var _old_bullets_spawn_interval: float
var _old_damage: float
var _paused_shots := false
var _active_fire := false
var _is_ready := false

onready var _shot: AudioStreamPlayer2D = $Shot
onready var _crash: AudioStreamPlayer2D = $Crash
onready var _shape: CollisionShape2D = $CollisionShape2D
onready var _current_sprite: Sprite = $Sprite
onready var _timer_boost: Timer = $TimerBoost


func _ready() -> void:
	hp.connect("hp_change", self, "_on_hp_change")
	get_tree().get_root().connect("size_changed", self, "_resize")
	_resize()
	_is_ready = true


func _physics_process(delta: float) -> void:
	if _active_fire and visible:
		_handle_gun(delta)


func _input(event: InputEvent) -> void:
	if event is InputEventMouseButton:
		if event.is_pressed():
			_active_fire = true
		else:
			_active_fire = false


func _notification(what) -> void:
	if what != NOTIFICATION_PREDELETE:
		return
	hp.free()


func _exit_tree() -> void:
	_timer_boost.paused = true


func _enter_tree() -> void:
	if not _is_ready:
		return
	_timer_boost.paused = false


func set_variant(variant, repair: bool) -> void:
	if is_boosted():
		reset_boost()
	var variants: Variants = Variants.new()
	var vars: Dictionary = variants.get_vars(variant)
	_rotate_speed = vars.rotate_speed
	_bullets_spawn_interval = vars.bullets_spawn_interval
	_bullets_variant = vars.bullets_variant
	_damage = vars.damage
	_fragments_variant = vars.fragments_variant
	_fragments_min = vars.fragments_min
	_fragments_max = vars.fragments_max
	_old_rotate_speed = _rotate_speed
	_old_bullets_spawn_interval = _bullets_spawn_interval
	_old_damage = _damage
	hp.set_max(vars.hp)
	if _current_type != vars.type or repair:
		hp.set_hp(vars.hp)
	_set_texture(vars.texture)
	_current_variant = variant
	_current_type = vars.type
	variants.free()


func get_damage() -> float:
	return _damage


func play_shot() -> void:
	_shot.play()


func hide_gun() -> void:
	hide()
	_shape.set_deferred("disabled", true)


func show_gun() -> void:
	show()
	_shape.set_deferred("disabled", false)


func get_variant():
	return _current_variant


func set_boost(boost: int, timeout: float) -> void:
	_old_rotate_speed = _rotate_speed
	_old_bullets_spawn_interval = _bullets_spawn_interval
	_old_damage = _damage
	_rotate_speed *= boost
	_bullets_spawn_interval /= boost
	_damage *= boost
	_timer_boost.start(timeout)


func reset_boost() -> void:
	_timer_boost.stop()
	_rotate_speed = _old_rotate_speed
	_bullets_spawn_interval = _old_bullets_spawn_interval
	_damage = _old_damage
	emit_signal("on_reset_boost", Actions.ActionType.BOOST)


func is_boosted() -> bool:
	return !_timer_boost.is_stopped()


func set_paused_shots(paused_shots: bool) -> void:
	_paused_shots = paused_shots


func centered_gun() -> void:
	var height: float = _HEIGHT
	position.y = -(get_viewport().size.y / 2 - height * 1.5)


func _resize() -> void:
	if is_inside_tree():
		centered_gun()


func _handle_gun(delta: float) -> void:
	# rotate gun
	var angle: float = get_angle_to(get_global_mouse_position()) + rotation
	if angle > rotation:
		rotation += _rotate_speed * delta
		if rotation > angle:
			rotation = angle
	else:
		rotation -= _rotate_speed * delta
		if rotation < angle:
			rotation = angle
	# handle fire
	_bullets_spawn_elasped += delta
	if _bullets_spawn_elasped >= _bullets_spawn_interval and not _paused_shots:
		var data := {
			"variant": _bullets_variant,
			"radius": _RADIUS,
			"rotation": rotation,
			"position": position,
		}
		emit_signal("spawn_bullet", data)
		_bullets_spawn_elasped = 0.0
		play_shot()
		Global.counter.increase(GameCounter.CountType.SHOTS, 1)


func _on_hp_change(new_hp: float, old_hp: float) -> void:
	match _current_type:
		Variants.Type.STANDARD:
			var part: float = hp.calc_part(4)
			if part == 4:
				_set_texture("standard")
			elif part >= 3:
				_set_texture("standard_damaged")
			elif part >= 2:
				_set_texture("standard_damaged_1")
			else:
				_set_texture("standard_damaged_2")
		Variants.Type.FLAT:
			var part: float = hp.calc_part(3)
			if part == 3:
				_set_texture("flat")
			elif part >= 2:
				_set_texture("flat_damaged")
			else:
				_set_texture("flat_damaged_1")
		Variants.Type.RED:
			var part: float = hp.calc_part(9)
			if part == 9:
				_set_texture("red")
			elif part >= 8:
				_set_texture("red_damaged")
			elif part >= 7:
				_set_texture("red_damaged_1")
			elif part >= 6:
				_set_texture("red_damaged_2")
			elif part >= 5:
				_set_texture("red_damaged_3")
			elif part >= 4:
				_set_texture("red_damaged_4")
			elif part >= 3:
				_set_texture("red_damaged_5")
			elif part >= 2:
				_set_texture("red_damaged_6")
			else:
				_set_texture("red_damaged_7")
		Variants.Type.BLUE:
			var part: float = hp.calc_part(5)
			if part == 5:
				_set_texture("blue")
			elif part >= 4:
				_set_texture("blue_damaged")
			elif part >= 3:
				_set_texture("blue_damaged_1")
			elif part >= 2:
				_set_texture("blue_damaged_2")
			else:
				_set_texture("blue_damaged_3")
	if new_hp < old_hp and new_hp != hp.maximum:
		var data := {
				"fragments_range": { "min": _fragments_min, "max": _fragments_max },
				"fragments_variant": _fragments_variant,
				"radius": _RADIUS,
				"position": position,
		}
		if new_hp == 0:
			hide_gun()
			emit_signal("on_die")
			Global.counter.increase(GameCounter.CountType.DEADS, 1)
		else:
			emit_signal("on_damaged", (1 / hp.maximum) / (1 / hp.current))
		emit_signal("on_fragments", data)
		_crash.play()


func _set_texture(name: String) -> void:
	var texture: Texture = _current_sprite.texture
	match name:
		"standard":
			texture = load("res://entity/gun/png/gun.png")
		"flat":
			texture = load("res://entity/gun/png/gun1.png")
		"red":
			texture = load("res://entity/gun/png/gun2.png")
		"blue":
			texture = load("res://entity/gun/png/gun3.png")
		"standard_damaged":
			texture = load("res://entity/gun/png/gun-damaged.png")
		"standard_damaged_1":
			texture = load("res://entity/gun/png/gun-damaged1.png")
		"standard_damaged_2":
			texture = load("res://entity/gun/png/gun-damaged2.png")
		"flat_damaged":
			texture = load("res://entity/gun/png/gun1-damaged.png")
		"flat_damaged_1":
			texture = load("res://entity/gun/png/gun1-damaged1.png")
		"red_damaged":
			texture = load("res://entity/gun/png/gun2-damaged.png")
		"red_damaged_1":
			texture = load("res://entity/gun/png/gun2-damaged1.png")
		"red_damaged_2":
			texture = load("res://entity/gun/png/gun2-damaged2.png")
		"red_damaged_3":
			texture = load("res://entity/gun/png/gun2-damaged3.png")
		"red_damaged_4":
			texture = load("res://entity/gun/png/gun2-damaged4.png")
		"red_damaged_5":
			texture = load("res://entity/gun/png/gun2-damaged5.png")
		"red_damaged_6":
			texture = load("res://entity/gun/png/gun2-damaged6.png")
		"red_damaged_7":
			texture = load("res://entity/gun/png/gun2-damaged7.png")
		"blue_damaged":
			texture = load("res://entity/gun/png/gun3-damaged.png")
		"blue_damaged_1":
			texture = load("res://entity/gun/png/gun3-damaged1.png")
		"blue_damaged_2":
			texture = load("res://entity/gun/png/gun3-damaged2.png")
		"blue_damaged_3":
			texture = load("res://entity/gun/png/gun3-damaged3.png")
	_current_sprite.texture = texture


func _on_TimerBoost_timeout() -> void:
	reset_boost()
