class_name Game
extends Node2D


const Cheats = preload("res://game/gd/cheats.gd")
const MultiTimer = preload("res://game/gd/multi_timer.gd")
const MusicFader = preload("res://game/gd/music_fader.gd")
const FIRST_LEVEL = Levels.Level.NEWBIE
const FIRST_ZEN_LEVEL = Levels.Level.ZEN_START
const ZEN_ENDLESS_LEVEL = Levels.Level.ZEN_ENDLESS

const _TRACK_MAX := 3

var _random: RandomNumberGenerator = RandomNumberGenerator.new()
var _levels: Levels = Levels.new()
var _cheats: Cheats = Cheats.new()
var _events: Events = Events.new()
var _multi_timer: MultiTimer = MultiTimer.new()
var _music_fader: MusicFader = MusicFader.new()
var _current_level
var _next_level
var _level_gun_variant
var _old_gun_variant
var _fortune_coef: float
var _duration_level := 0.0
var _duration_wait := 0.0
var _bg_track_number := 0
var _meteors_random_interval := 0.25
var _meteors_random_amount := 0
var _action_boost := 0
var _action_boost_timeout := 0.0
var _action_border_timeout := 0.0
var _current_random_event := -1
var _game_over := false
var _waiting_level := false
var _wait_clear := false
var _level_clear := false
var _end_game := false
var _cheat_play := false
var _show_donate := false
var _show_review := false
var _is_ready := false

onready var gun: Gun = $Gun
onready var _debug_container: VBoxContainer = $CanvasLayer/Debug
onready var _bg_music: AudioStreamPlayer2D = $Bg
onready var _bg_end_music: AudioStreamPlayer2D = $BgEnd
onready var _sound_activate: AudioStreamPlayer2D = $SoundActivate
onready var _fail: AudioStreamPlayer2D = $Fail
onready var _sound_splash: AudioStreamPlayer2D = $SplashSound
onready var _warning: Warning = $Warning
onready var _splash_sprite: AnimatedSprite = $SplashSprite
onready var _pause_panel: Panel = $CanvasLayer/PausePanel
onready var _pause_button: TextureButton = $CanvasLayer/PauseButton
onready var _actions: Actions = $Actions
onready var _border: Border = $Border
onready var _drops: Drops = $Drops
onready var _gifts: Gifts = $Gifts
onready var _meteors: Meteors = $Meteors
onready var _fragments: Fragments = $Fragments
onready var _bullets: Bullets = $Bullets
onready var _timer_gun: Timer = $TimerGun
onready var _timer_wait: Timer = $TimerWait
onready var _timer_level: Timer = $TimerLevel
onready var _timer_meteors: Timer = $TimerMeteors
onready var _timer_events: Timer = $TimerEvents
onready var _end_game_menu: VBoxContainer = $CanvasLayer/EndGame
onready var _hearts: Rain = $Rain
onready var _salute: Salute = $Salute


func _ready() -> void:
	_random.randomize()
	set_level(FIRST_LEVEL)
	gun.connect("on_die", self, "_on_gun_die")
	gun.connect("on_fragments", _fragments, "spawn_fragments")
	gun.connect("on_reset_boost", self, "_action_end")
	gun.connect("spawn_bullet", _bullets, "spawn_bullet")
	gun.connect("on_damaged", self, "_on_gun_damaged")
	_cheats.connect("cheat_combination", self, "_cheat_activate")
	_actions.setup_game(self)
	_border.connect("on_deactivate", self, "_action_end")
	_drops.connect("on_end_rain", self, "_action_end")
	_gifts.connect("gift_activate", self, "_gift_activate")
	_meteors.connect("end_random_bang", self, "_action_end")
	_meteors.connect("spawn_fragments", _fragments, "spawn_fragments")
	_meteors.connect("on_empty", self, "_meteors_empty")
	_multi_timer.connect("on_timeout", self, "_event_activate")
	_is_ready = true


func _notification(what) -> void:
	if what != NOTIFICATION_PREDELETE:
		return
	_levels.free()
	_cheats.free()
	_events.free()
	_multi_timer.free()


func _exit_tree() -> void:
	_timer_events.paused = true
	_timer_gun.paused = true
	_timer_level.paused = true
	_timer_meteors.paused = true
	_timer_wait.paused = true
	_bg_end_music.stop()


func _enter_tree() -> void:
	if not _is_ready:
		return
	if _waiting_level:
		_bg_music.call_deferred("stop")
	_timer_events.paused = false
	_timer_gun.paused = false
	_timer_level.paused = false
	_timer_meteors.paused = false
	_timer_wait.paused = false
	if _salute.active:
		_bg_end_music.play()


func _process(delta: float) -> void:
	if _waiting_level and not _timer_wait.is_stopped():
		_music_fader.tick(delta, _duration_wait)
	if not _game_over:
		_multi_timer.tick(delta)


func _input(event: InputEvent) -> void:
	if event is InputEventMouseButton and not event.is_pressed():
		_cheats.touch(get_viewport().size, get_global_mouse_position())


#
# LEVELS
#

func set_level(level) -> void:
	var vars: Dictionary = _levels.get_level(level)
	_duration_level = vars.time
	_duration_wait = vars.time_wait
	_current_level = level
	_next_level = vars.next_level
	_level_gun_variant = vars.gun_variant
	_fortune_coef = vars.fortune_coef
	_wait_clear = vars.wait_clear
	_end_game = vars.end_game
	_cheat_play = vars.cheat_play
	_show_donate = vars.show_donate
	_show_review = vars.show_review
	_meteors.set_variants(vars.meteors)
	gun.set_variant(vars.gun_variant, vars.gun_repair)
	gun.set_paused_shots(false)
	gun.show_gun()
	_music_fader.reset()
	set_bg_music(vars.track)
	_bg_music.play()
	_start_level_timer()
	_actions.hide_all()
	_meteors_random_amount = vars.action_random.amount
	_meteors_random_interval = vars.action_random.interval
	if vars.action_random.enabled:
		_actions.call_deferred(
				"set_animation_speed",
				Actions.ActionType.RANDOM, vars.action_random.speed
		)
		_actions.set_action_visible(Actions.ActionType.RANDOM, true)
	_action_boost = vars.action_boost.boost
	_action_boost_timeout = vars.action_boost.timeout
	if vars.action_boost.enabled:
		_actions.call_deferred(
				"set_animation_speed",
				Actions.ActionType.BOOST, vars.action_boost.speed
		)
		_actions.set_boost_animation(vars.action_boost.boost)
		_actions.set_action_visible(Actions.ActionType.BOOST, true)
	_action_border_timeout = vars.action_border.timeout
	if vars.action_border.enabled:
		_actions.call_deferred(
				"set_animation_speed",
				Actions.ActionType.BORDER, vars.action_border.speed
		)
		_actions.set_action_visible(Actions.ActionType.BORDER, true)
	_drops.set_amount(vars.action_drop.amount)
	_drops.set_drop_variants(vars.action_drop.variants)
	_drops.set_interval(vars.action_drop.interval_min, vars.action_drop.interval_max)
	if vars.action_drop.enabled:
		_actions.call_deferred(
				"set_animation_speed",
				Actions.ActionType.DROP, vars.action_drop.speed
		)
		_actions.set_drop_animation(vars.action_drop.animation)
		_actions.set_action_visible(Actions.ActionType.DROP, true)
	for time in vars.events:
		if (
				vars.random.enabled
				and vars.random.events.mode != Levels.RndEventsMode.NONE
				and vars.random.events.list.has(time)
		):
			continue
		var event: Dictionary = _events.get_vars(vars.events[time])
		_multi_timer.add_task(time, event)
	if vars.random.enabled:
		if vars.random.meteors.mode != Levels.RndMeteorsMode.NONE:
			_set_random_meteors(vars)
			_start_meteors_timer(vars)
		if vars.random.events.mode != Levels.RndEventsMode.NONE:
			_setup_random_event(vars)
	if vars.hearts:
		_hearts.spawn_pause = rand_range(20.0, 30.0)
	_hearts.set_process(vars.hearts)
	if vars.endless_splash:
		_splash_sprite.play_animation()
		_sound_splash.play()


func _start_next_level_timer():
	_timer_meteors.stop()
	_timer_events.stop()
	if _show_donate:
		Global.donate.show_donate_button()
	if _show_review:
		Global.review.show_review_button()
	if _duration_wait == 0:
		_multi_timer.clear()
		set_level(_next_level)
		return
	_waiting_level = true
	_meteors.set_pause_spawn(true)
	_meteors.stop_wave()
	_timer_wait.start(_duration_wait)
	if _cheat_play:
		_salute.play_cheat(_cheats.ZEN_ENDLESS_LEVEL)


func _start_level_timer():
	if _duration_level <= 0:
		return
	_timer_level.stop()
	_timer_level.start(_duration_level)


func _set_next_level():
	_multi_timer.clear()
	_bg_music.stop()
	if _wait_clear and not _level_clear:
		return
	_level_clear = false
	_waiting_level = false
	_meteors.set_pause_spawn(false)
	_meteors.stop_random_bang()
	_meteors.stop_wave()
	set_level(_next_level)


func _set_first_level(zen:=false) -> void:
	stop_game()
	if zen:
		set_level(FIRST_ZEN_LEVEL)
	else:
		set_level(FIRST_LEVEL)
		Global.counter.increase(GameCounter.CountType.GAME_RESTARTS, 1)


func _meteors_empty(is_empty: bool) -> void:
	_level_clear = is_empty
	gun.set_paused_shots(is_empty)
	if _end_game:
		_to_end_game()
	elif _timer_wait.is_stopped() and _wait_clear:
		_set_next_level()


#
# GAME
#

func stop_game() -> void:
	_game_over = false
	_actions.visible = true
	_level_clear = false
	_waiting_level = false
	_salute.active = false
	gun.reset_boost()
	gun.set_paused_shots(false)
	_timer_events.stop()
	_timer_meteors.stop()
	_timer_gun.stop()
	_timer_level.stop()
	_timer_wait.stop()
	_actions.reset()
	_actions.hide_all()
	_multi_timer.clear()
	_border.force_deactivate_border()
	_drops.stop()
	_drops.clear()
	_gifts.clear()
	_meteors.set_pause_spawn(false)
	_meteors.stop_random_bang()
	_meteors.stop_wave()
	_meteors.clear()
	_bullets.clear()
	_bg_end_music.stop()
	_warning.stop()


func _to_end_game() -> void:
	stop_game()
	gun.hide_gun()
	get_tree().paused = true
	_salute.active = true
	_bg_music.stop()
	_music_fader.reset()
	_bg_end_music.play()
	_pause_button.hide()
	_end_game_menu.show()


#
# BG MUSIC
#

func set_bg_music(num_track: int) -> void:
	if num_track < 0 or num_track > _TRACK_MAX or num_track == _bg_track_number:
		return
	var stream: AudioStream = load("res://game/ogg/" + str(num_track) + ".ogg")
	_bg_music.stop()
	_bg_music.stream = stream
	_bg_music.play()
	_bg_track_number = num_track


#
# ACTIONS
#

func _cheat_activate(name: String):
	print("cheat: ", name)
	match(name):
		"debug_info":
			_debug_container.visible = !_debug_container.visible
		"next_bg_music":
			var num_track: int = _bg_track_number + 1
			if num_track > _TRACK_MAX:
				num_track = 0
			set_bg_music(num_track)
		"reset_actions":
			_actions.reset()
		"game_over":
			gun.hp.current -= gun.hp.current
		"test_level":
			stop_game()
			set_level(Levels.Level.TEST)
		"zen_level":
			stop_game()
			set_level(FIRST_ZEN_LEVEL)
		"zen_endless_level":
			stop_game()
			set_level(ZEN_ENDLESS_LEVEL)


func action_activate(name: String, gift:=false) -> void:
	if _game_over:
		_actions.reset()
		return
	match(name):
		"random":
			if _meteors_random_amount > 0 and not _meteors.is_random_bang_active():
				_meteors.bang_random_meteors(_meteors_random_amount, _meteors_random_interval)
				_actions.set_enabled(Actions.ActionType.RANDOM, false)
		"boost":
			if _action_boost > 0 and not gun.is_boosted():
				gun.set_boost(_action_boost, _action_boost_timeout)
				_actions.set_enabled(Actions.ActionType.BOOST, false)
		"border":
			if not _border.visible:
				_border.activate_border(_action_border_timeout)
				_actions.set_enabled(Actions.ActionType.BORDER, false)
		"drop":
			if not _drops.is_active():
				_drops.start()
				_actions.set_enabled(Actions.ActionType.DROP, false)
	if not gift:
		if _fortune_coef >= _random.randf():
			_gifts.spawn_gift(Gift.GiftVariant.RENEWAL)
		Global.counter.increase(GameCounter.CountType.ACTIONS, 1)


func _action_end(action) -> void:
	match(action):
		Actions.ActionType.RANDOM:
			_actions.set_enabled(Actions.ActionType.RANDOM, true)
			if not _waiting_level and _fortune_coef >= _random.randf():
				_gifts.spawn_gift(Gift.GiftVariant.RANDOM)
		Actions.ActionType.BOOST:
			_actions.set_enabled(Actions.ActionType.BOOST, true)
			if not _waiting_level and _fortune_coef >= _random.randf():
				_gifts.spawn_gift(Gift.GiftVariant.BOOST)
		Actions.ActionType.BORDER:
			_actions.set_enabled(Actions.ActionType.BORDER, true)
			if not _waiting_level and _fortune_coef >= _random.randf():
				_gifts.spawn_gift(Gift.GiftVariant.BORDER)
		Actions.ActionType.DROP:
			_actions.set_enabled(Actions.ActionType.DROP, true)
			if not _waiting_level and _fortune_coef >= _random.randf():
				_gifts.spawn_gift(Gift.GiftVariant.DROP)


func _gift_activate(variant, to: Node) -> void:
	if to.is_in_group("bullet"):
		to.bang()
	match(variant):
		Gift.GiftVariant.BOOST:
			action_activate("boost", true)
		Gift.GiftVariant.BORDER:
			action_activate("border", true)
		Gift.GiftVariant.DROP:
			action_activate("drop", true)
		Gift.GiftVariant.RANDOM:
			action_activate("random", true)
		Gift.GiftVariant.REPAIR:
			gun.hp.set_hp(gun.hp.get_max())
		Gift.GiftVariant.RENEWAL:
			_actions.reset()
	Global.counter.increase(GameCounter.CountType.GIFTS_ACTIVATED, 1)


#
# EVENTS
#

func _event_activate(event: Dictionary) -> void:
	match(event.type):
		Events.EventType.WARNING:
			_warning.start_play(event.timeout)
		Events.EventType.GIFT:
			if event.chance >= _random.randf():
				_gifts.spawn_gift(event.variant)
		Events.EventType.METEORS_WAVE:
			_meteors.start_wave(event.meteors, event.timeout_min, event.timeout_max)
		Events.EventType.GUN_UPGRADE:
			if not _timer_gun.is_stopped():
				return
			_sound_activate.play()
			if event.timeout > 0:
				_old_gun_variant = gun.get_variant()
				_timer_gun.start(event.timeout)
			gun.set_variant(event.gun_variant, event.gun_repair)


func _on_TimerGun_timeout() -> void:
	gun.set_variant(_old_gun_variant, false)


#
# RANDOM
#

func _on_TimerMeteors_timeout() -> void:
	var vars: Dictionary = _levels.get_level(_current_level)
	match(vars.random.meteors.mode):
		Levels.RndMeteorsMode.ALL:
			var random_meteors := []
			for _i in _meteors.get_variants().size() - 1:
				var index: int = _random.randi_range(0, vars.meteors.size() - 1)
				random_meteors.append(vars.meteors[index])
			_meteors.set_variants(random_meteors)
		Levels.RndMeteorsMode.RANDOM:
			_set_random_meteors(vars)
		Levels.RndMeteorsMode.SOME:
			var random_meteors: Array = _meteors.get_variants()
			random_meteors.remove(_random.randi_range(0, random_meteors.size() - 1))
			var index: int = _random.randi_range(0, vars.meteors.size() - 1)
			random_meteors.append(vars.meteors[index])
			_meteors.set_variants(random_meteors)
	_start_meteors_timer(vars)


func _on_TimerEvents_timeout() -> void:
	_setup_random_event(_levels.get_level(_current_level))


func _start_meteors_timer(vars: Dictionary) -> void:
	var interval: float = _random.randf_range(
			vars.random.meteors.change_interval.min,
			vars.random.meteors.change_interval.max
	)
	_timer_meteors.start(interval)


func _set_random_meteors(vars: Dictionary) -> void:
	var random_meteors := []
	var indexes := []
	for _i in _random.randi_range(
			vars.random.meteors.variants.min,
			vars.random.meteors.variants.max
	):
		var index: int = _random.randi_range(0, vars.meteors.size() - 1)
		while indexes.has(index):
			index = _random.randi_range(0, vars.meteors.size() - 1)
		random_meteors.append(vars.meteors[index])
		indexes.append(index)
	_meteors.set_variants(random_meteors)


func _setup_random_event(vars: Dictionary) -> void:
	if not _timer_events.is_stopped():
		return
	var time: float = _random.randf_range(
			vars.random.events.interval.min,
			vars.random.events.interval.max
	)
	match(vars.random.events.mode):
		Levels.RndEventsMode.ORDER:
			if _current_random_event == -1:
				var event: Dictionary = _events.get_vars(vars.events[vars.random.events.list[0]])
				_multi_timer.add_task(time, event)
				_current_random_event = 0
			else:
				_current_random_event += 1
				if _current_random_event == vars.random.events.list.size():
					_current_random_event = 0
				var event_key = vars.random.events.list[_current_random_event]
				var event: Dictionary = _events.get_vars(vars.events[event_key])
				_multi_timer.add_task(time, event)
		Levels.RndEventsMode.RANDOM:
			var index: int = _random.randi_range(
					0,
					vars.random.events.list.size() - 1
			)
			var event: Dictionary = _events.get_vars(vars.events[vars.random.events.list[index]])
			_multi_timer.add_task(time, event)
	_timer_events.start(time)


#
# GUN
#

func _on_gun_die() -> void:
	if not _timer_gun.is_stopped():
		gun.set_variant(_old_gun_variant, false)
	_game_over = true
	_meteors.set_pause_spawn(true)
	_timer_events.stop()
	_timer_meteors.stop()
	_timer_gun.stop()
	_timer_level.stop()
	_timer_wait.stop()
	_actions.visible = false
	_fail.play()


func _on_gun_damaged(hp_coef: float) -> void:
	if _fortune_coef <= 0:
		return
	var chance: float = _fortune_coef / hp_coef
	if chance >= _random.randf():
		_gifts.spawn_gift(Gift.GiftVariant.REPAIR)


#
# PAUSE MENU
#

func _on_ToMenuButton_pressed() -> void:
	if _waiting_level:
		_music_fader.reset()
	Global.live_change_scene("res://menu/main_menu.tscn")


func _on_RestartButton_pressed() -> void:
	stop_game()
	set_level(_current_level)
	_on_ResumeButton_pressed()
	Global.counter.increase(GameCounter.CountType.LEVEL_RESTARTS, 1)


func _on_GameRestartButton_pressed() -> void:
	_set_first_level()
	_on_ResumeButton_pressed()


func _on_ResumeButton_pressed() -> void:
	_pause_panel.hide()
	get_tree().paused = false
	get_tree().get_root().emit_signal("size_changed")


func _on_PauseButton_pressed() -> void:
	get_tree().paused = true
	_pause_panel.show()


#
# END GAME MENU
#

func _on_ZenButton_pressed() -> void:
	get_tree().paused = false
	_pause_button.show()
	_end_game_menu.hide()
	_set_first_level(true)
	get_tree().get_root().emit_signal("size_changed")


func _on_RestartGameButton_pressed():
	get_tree().paused = false
	_pause_button.show()
	_end_game_menu.hide()
	_set_first_level()
	get_tree().get_root().emit_signal("size_changed")


func _on_StatisticButton_pressed() -> void:
	_music_fader.reset()
	Global.live_change_scene("res://menu/statistic.tscn")


func _on_MenuButton_pressed() -> void:
	_music_fader.reset()
	Global.live_change_scene("res://menu/main_menu.tscn")
