class_name Drops
extends Node2D


signal on_end_rain

const DROP = preload("res://entity/drop/drop.tscn")
const Postponer = preload("res://game/gd/postponer.gd")
const DropList = preload("res://game/gd/drop_list.gd")

var _postponer: Postponer = Postponer.new()
var _drop_list: DropList = DropList.new()
var _drop_variants := {
	"yellow": { "amount": 20, "impulse": 400.0 },
	"red": { "amount": 5, "impulse": 400.0 },
	"blue": { "amount": 5, "impulse": 400.0 },
}
var _interval_min: float
var _interval_max: float
var _amount: int
var _count := 0
var _active := false

onready var _timer: Timer = $Timer


func _ready() -> void:
	randomize()


func _process(_delta: float) -> void:
	if _postponer.tick():
		_handle_drops()


func _notification(what) -> void:
	if what != NOTIFICATION_PREDELETE:
		return
	_postponer.free()
	_drop_list.free()


func set_interval(i_min: float, i_max: float) -> void:
	_interval_min = i_min
	_interval_max = i_max


func set_amount(amount: int) -> void:
	_amount = amount


func set_drop_variants(drop_variants: Dictionary) -> void:
	_drop_variants = drop_variants


func start() -> void:
	_count = 1
	_active = true
	_drop_list.recount()
	call_deferred("_spawn_drop")
	_timer.start(rand_range(_interval_min, _interval_max))


func stop() -> void:
	_active = false
	if not _timer.is_stopped():
		_timer.stop()
	emit_signal("on_end_rain", Actions.ActionType.DROP)


func clear() -> void:
	_drop_list.iterate(self, "_clear_handler")
	_drop_list.clear()


func is_active() -> bool:
	return _active


func _clear_handler(drop: Drop) -> void:
	call_deferred("remove_child", drop)
	drop.queue_free()


func _spawn_drop() -> void:
	var variant = _drop_list.get_next_variant(
			_drop_variants.yellow.amount,
			_drop_variants.red.amount,
			_drop_variants.blue.amount
	)
	if variant == Drop.DropVariant.NONE:
		return
	var drop: Drop = DROP.instance()
	drop.set_variant(variant)
	add_child(drop)
	var size: Vector2 = drop.get_sprite_size()
	var limit_x: float = get_viewport().size.x / 2 - size.x * 1.2
	drop.position.x = rand_range(-limit_x, limit_x)
	drop.position.y -= get_viewport().size.y / 2 + size.y
	var point := Vector2(drop.position.x, get_viewport().size.y / 2)
	var impulse_direction: Vector2 = drop.position.direction_to(point)
	var impulse_strength: float = _drop_variants[drop.get_string_variant()].impulse
	impulse_direction.x *= impulse_strength
	impulse_direction.y *= impulse_strength
	drop.apply_central_impulse(impulse_direction)
	drop.connect("on_die", self, "_remove_drop")
	_drop_list.add_drop(drop)


func _handle_drops(drop: Drop = null) -> void:
	if not drop:
		_drop_list.iterate(self, "_handle_drops")
		return
	var size: Vector2 = drop.get_sprite_size()
	var limit_x: float = get_viewport().size.x / 2 + size.x
	var limit_y: float = get_viewport().size.y / 2 + size.y * 1.5
	if (
			drop.position.x > limit_x or drop.position.x < -limit_x
			or drop.position.y > limit_y or drop.position.y < -limit_y
	):
		_remove_drop(drop)


func _remove_drop(drop: Drop) -> void:
	_drop_list.remove_drop(drop)
	drop.queue_free()


func _on_Timer_timeout() -> void:
	if not _active:
		return
	_count += 1
	if _count >= _amount:
		_active = false
		emit_signal("on_end_rain", Actions.ActionType.DROP)
		return
	call_deferred("_spawn_drop")
	_timer.start(rand_range(_interval_min, _interval_max))
