class_name Meteors
extends Node2D


signal end_random_bang
signal spawn_fragments
signal on_empty

const METEOR = preload("res://entity/meteor/meteor.tscn")
const Postponer = preload("res://game/gd/postponer.gd")

var _random := RandomNumberGenerator.new()
var _postponer: Postponer = Postponer.new()
var _wave_counter := {}
var _meteors := []
var _meteors_variants: Array
var _wave_meteors_variants: Array
var _wave_timeout_min: float
var _wave_timeout_max: float
var _meteors_random_amount := 0
var _meteors_bang_counter := 0
var _pause_spawn := false

onready var _timer_random: Timer = $TimerRandom
onready var _timer_wave: Timer = $TimerWave


func _ready() -> void:
	_random.randomize()
	_spawn_meteors()


func _process(_delta: float) -> void:
	if _postponer.tick():
		_handle_meteors()


func _notification(what) -> void:
	if what != NOTIFICATION_PREDELETE:
		return
	_postponer.free()


func set_variants(variants: Array) -> void:
	_meteors_variants = variants


func get_variants() -> Array:
	return _meteors_variants


func set_pause_spawn(pause_spawn: bool) -> void:
	_pause_spawn = pause_spawn


func clear() -> void:
	for meteor in _meteors:
		call_deferred("remove_child", meteor)
		meteor.queue_free()
	_meteors.clear()
	_wave_counter.clear()


func bang_random_meteors(amount: int, interval: float) -> void:
	_meteors_bang_counter = 0
	_meteors_random_amount = amount
	_timer_random.start(interval)
	_bang_random_meteor()


func start_wave(variants, timeout_min, timeout_max) -> void:
	_wave_meteors_variants = variants
	_wave_timeout_min = timeout_min
	_wave_timeout_max = timeout_max
	_wave_counter.clear()
	_spawn_wave_meteor()
	_timer_wave.start(_random.randf_range(timeout_min, timeout_max))


func stop_random_bang() -> void:
	emit_signal("end_random_bang", Actions.ActionType.RANDOM)
	_timer_random.stop()


func stop_wave() -> void:
	_timer_wave.stop()


func is_random_bang_active() -> bool:
	return !_timer_random.is_stopped()


func _on_meteor_collide(meteor: Meteor, to: Node) -> void:
	if (to.is_in_group("bullet") and to.is_active()) or to.is_in_group("hp"):
		meteor.hp.current -= to.get_damage()
		to.hp.current -= meteor.get_damage()
		if to.is_in_group("bullet") and to.hp.current == 0:
			Global.counter.increase(GameCounter.CountType.SHOTS_METEORS, 1)


func _on_meteor_die(meteor: Meteor) -> void:
	_remove_meteor(meteor)


func _remove_meteor(meteor: Meteor) -> void:
	call_deferred("remove_child", meteor)
	_meteors.erase(meteor)
	meteor.queue_free()
	if _pause_spawn and _meteors.empty():
		emit_signal("on_empty", true)


func _spawn_meteor(variant) -> void:
	var meteor: Meteor = METEOR.instance()
	add_child(meteor)
	meteor.set_variant(variant)
	var size: Vector2 = meteor.get_sprite_size()
	meteor.position.y = get_viewport().size.y / 2 + size.y
	var limit_x: float = get_viewport().size.x / 2 - size.x * 1.2
	meteor.position.x = _random.randf_range(-limit_x, limit_x)
	var impulse_range: Dictionary = meteor.get_impulse_range()
	var impulse_strength: float = _random.randf_range(impulse_range.min, impulse_range.max)
	var point := Vector2(
		_random.randf_range(-limit_x, limit_x),
		-(get_viewport().size.y / 2)
	)
	var impulse_direction: Vector2 = meteor.position.direction_to(point)
	impulse_direction.x *= impulse_strength
	impulse_direction.y *= impulse_strength
	meteor.apply_central_impulse(impulse_direction)
	meteor.connect("on_collide", self, "_on_meteor_collide")
	meteor.connect("on_die", self, "_on_meteor_die")
	meteor.connect("on_fragments", self, "_spawn_fragments")
	_meteors.append(meteor)
	Global.counter.increase(GameCounter.CountType.METEORS_SPAWNED, 1)


func _spawn_meteors() -> void:
	var variants_count := {}
	for meteor in _meteors:
		if variants_count.has(meteor.get_variant()):
			variants_count[meteor.get_variant()] += 1
		else:
			variants_count[meteor.get_variant()] = 1
	for variant in _meteors_variants:
		var enum_var = variant.variant
		var var_count: int = variants_count[enum_var] if variants_count.has(enum_var) else 0
		for _i in range(variant.count - var_count):
			_spawn_meteor(enum_var)


func _spawn_wave_meteor() -> bool:
	for variant in _wave_meteors_variants:
		if not _wave_counter.has(variant.variant):
			_wave_counter[variant.variant] = 1
			_spawn_meteor(variant.variant)
			return true
		elif _wave_counter[variant.variant] < variant.count:
			_wave_counter[variant.variant] += 1
			_spawn_meteor(variant.variant)
			return true
	_wave_counter.clear()
	return false


func _handle_meteors() -> void:
	var for_remove := []
	var amount_passed := 0
	for meteor in _meteors:
		var size: Vector2 = meteor.get_sprite_size()
		var limit_x: float = get_viewport().size.x / 2 + size.x
		var limit_y: float = get_viewport().size.y / 2 + size.y * 1.5
		if (
				meteor.position.x > limit_x or meteor.position.x < -limit_x
				or meteor.position.y > limit_y
		):
			for_remove.append(meteor)
		elif meteor.position.y < -limit_y:
			for_remove.append(meteor)
			amount_passed += 1
	Global.counter.increase(GameCounter.CountType.METEORS_PASSED, amount_passed)
	for meteor in for_remove:
		_remove_meteor(meteor)
	if not _pause_spawn:
		_spawn_meteors()


func _bang_random_meteor() -> void:
	var meteors_list := []
	for meteor in _meteors:
		var size: Vector2 = meteor.get_sprite_size()
		if meteor.position.y <= get_viewport().size.y / 2 + size.y * 0.5:
			meteors_list.append(meteor)
	if meteors_list.empty():
		return
	var index: int = _random.randi_range(0, meteors_list.size() - 1)
	meteors_list[index].bang()
	_meteors_bang_counter += 1
	if _meteors_bang_counter >= _meteors_random_amount:
		_timer_random.stop()
		emit_signal("end_random_bang", Actions.ActionType.RANDOM)


func _spawn_fragments(data: Dictionary) -> void:
	emit_signal("spawn_fragments", data)


func _on_TimerRandom_timeout() -> void:
	_bang_random_meteor()


func _on_TimerWave_timeout() -> void:
	if _spawn_wave_meteor():
		_timer_wave.start(_random.randf_range(_wave_timeout_min, _wave_timeout_max))
