extends Node


const Store = preload("res://global/gd/settings_store.gd")
const Counter = preload("res://global/gd/counter.gd")
const Donate = preload("res://global/gd/donate.gd")
const Review = preload("res://global/gd/review.gd")
const Letter = preload("res://global/gd/letter.gd")
const VERSION := 1.03
const AUTHORS := "snegg21"

const _MAX_WAIT_TIME := 100  # msec
const _MIN_WIDTH := 260
const _MIN_HEIGHT := 400

var store: Store = Store.new()
var counter: GameCounter = Counter.new()
var donate: Donate = Donate.new()
var review: Review = Review.new()
var letter: Letter = Letter.new()
var current_scene
var back_scene

var _loader: ResourceInteractiveLoader
var _wait_frames: int
var _pause_state := false

onready var _load_sprite: Sprite = $LoadSprite
onready var _rotate_sprite: Sprite = $CanvasLayer/RotateSprite
onready var _small_sprite: Sprite = $CanvasLayer/SmallSprite
onready var _rotate_label: Label = $CanvasLayer/RotateLabel
onready var _small_label: Label = $CanvasLayer/SmallLabel
onready var _canvas_modulate: CanvasModulate = $CanvasModulate
onready var _bg_music: AudioStreamPlayer2D = $MenuBg
onready var _request: HTTPRequest = $HTTPRequest


func _ready() -> void:
	get_tree().get_root().connect("size_changed", self, "_resize")
	OS.min_window_size = Vector2(_MIN_WIDTH, _MIN_HEIGHT)
	OS.max_window_size = Vector2(768, 1024)
	var root: Viewport = get_tree().get_root()
	current_scene = root.get_child(root.get_child_count() - 1)
	store.load_settings()
	_resize()


func _exit_tree() -> void:
	letter.save()
	store.free()
	counter.free()
	donate.free()
	review.free()
	letter.free()


func _process(_delta: float) -> void:
	if not _loader:
		set_process(false)
		return
	if _wait_frames > 0:
		_wait_frames -= 1
		return
	var time: int = OS.get_ticks_msec()
	while OS.get_ticks_msec() < time + _MAX_WAIT_TIME:
		var error := _loader.poll()
		if error == ERR_FILE_EOF:
			var resource: Resource = _loader.get_resource()
			_loader = null
			current_scene = resource.instance()
			get_tree().get_root().add_child(current_scene)
			get_tree().set_current_scene(current_scene)
			_load_sprite.hide()
			break
		elif error == OK:
			pass
		else:
			_loader = null
			break


func play_bg_music() -> void:
	if not _bg_music.playing:
		_bg_music.play()


func stop_bg_music() -> void:
	_bg_music.stop()


func http_request(url: String, data:=""):
	return _request.http_request(url, data)


func bg_goto_scene(path: String) -> void:
	_loader = ResourceLoader.load_interactive(path)
	if not _loader:
		return
	set_process(true)
	current_scene.queue_free()
	_load_sprite.show()
	_wait_frames = 60


func goto_scene(path: String) -> void:
	call_deferred("_deffered_goto_scene", path, false)


func live_change_scene(path: String) -> void:
	call_deferred("_deffered_goto_scene", path, true)


func return_scene() -> void:
	call_deferred("_deffered_return_scene")


func _resize() -> void:
	# centered
	var xc: float = get_viewport().size.x / 2
	var yc: float = get_viewport().size.y / 2
	_load_sprite.position.x = xc
	_load_sprite.position.y = yc
	_rotate_sprite.position.x = xc
	_rotate_sprite.position.y = yc
	_small_sprite.position.x = xc
	_small_sprite.position.y = yc
	_bg_music.position.x = xc
	_bg_music.position.y = yc
	# hide
	if _rotate_sprite.visible or _small_sprite.visible:
		get_tree().paused = _pause_state
	_rotate_sprite.hide()
	_rotate_label.hide()
	_small_sprite.hide()
	_small_label.hide()
	_canvas_modulate.hide()
	# check orientation
	if get_viewport().size.x > get_viewport().size.y:
		_rotate_sprite.show()
		_rotate_label.show()
		_canvas_modulate.show()
		_pause_state = get_tree().paused
		get_tree().paused = true
	# check size
	elif get_viewport().size.x < _MIN_WIDTH or get_viewport().size.y < _MIN_HEIGHT:
		_small_sprite.show()
		_small_label.show()
		_canvas_modulate.show()
		_pause_state = get_tree().paused
		get_tree().paused = true
	# logo resize
	if get_viewport().size.x < _load_sprite.texture.get_size().x:
		var scale_val: float = get_viewport().size.x * 0.0025
		_load_sprite.scale.x = scale_val
		_load_sprite.scale.y = scale_val
	else:
		_load_sprite.scale.x = 1
		_load_sprite.scale.y = 1


func _deffered_return_scene() -> void:
	current_scene.queue_free()
	current_scene = back_scene
	back_scene = null
	get_tree().get_root().add_child(current_scene)
	get_tree().set_current_scene(current_scene)
	current_scene.show()


func _deffered_goto_scene(path: String, pause: bool) -> void:
	if pause:
		back_scene = current_scene
		back_scene.hide()
		get_tree().get_root().remove_child(back_scene)
	else:
		current_scene.queue_free()
	var scene: Resource = ResourceLoader.load(path)
	current_scene = scene.instance()
	get_tree().get_root().add_child(current_scene)
	get_tree().set_current_scene(current_scene)


func _on_HTTPRequest_request_completed(
		result: int, response_code: int, headers: PoolStringArray, body: PoolByteArray
) -> void:
	donate.request_completed(result, response_code, headers, body)
	review.request_completed(result, response_code, headers, body)
